"""
Script to train xgboost models.
Right now only supports classification over nodes, without edges.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
from argparse import Namespace
from ..utils.read_nodes import read_all_nodes
from xgboost import XGBClassifier
from logging import Logger
import logging
from sklearn.model_selection import train_test_split
import numpy as np
from typing import Dict, Any, Tuple, Optional
from sklearn.model_selection import StratifiedKFold
import pandas as pd
from concurrent.futures import ThreadPoolExecutor
from ..utils.classification import metrics_from_predictions


def train(
        conf: Dict[str, Any],
        X_tr: np.ndarray,
        y_tr: np.ndarray,
        val_size: float,
        seed: int,
        num_classes: Optional[int] = 2
        ) -> XGBClassifier:
    X_tr, X_val, y_tr, y_val = train_test_split(
        X_tr, y_tr, test_size=val_size, stratify=y_tr, random_state=seed
    )
    if num_classes == 2:
        model = XGBClassifier(
            n_estimators=conf['n_estimators'],
            learning_rate=conf['learning_rate'],
            max_depth=conf['max_depth'],
            colsample_bytree=conf['colsample_bytree'],
            eval_metric='logloss',
            early_stopping_rounds=10
        )
    else:
        model = XGBClassifier(
            n_estimators=conf['n_estimators'],
            learning_rate=conf['learning_rate'],
            max_depth=conf['max_depth'],
            colsample_bytree=conf['colsample_bytree'],
            eval_metric='mlogloss',
            objective='multi:softmax',
            num_class=num_classes,
            early_stopping_rounds=10
        )
    model.fit(
        X_tr, y_tr,
        eval_set=[(X_val, y_val)], verbose=False
    )
    return model


def evaluate(
        model: XGBClassifier,
        X_val: np.ndarray,
        y_val: np.ndarray,
        num_classes: Optional[int] = 2
        ) -> Tuple[float, float, float]:
    preds = model.predict(X_val)
    if num_classes == 2:
        probs = model.predict_proba(X_val)[:, 1]
        acc, f1, auc, perc_err, ece = metrics_from_predictions(y_val, preds, probs, 2)
        return f1, acc, auc, perc_err, ece
    else:
        probs = model.predict_proba(X_val)
        micro, macro, weighted, ece = metrics_from_predictions(y_val, preds, probs, num_classes)
        return micro, macro, weighted, ece


def save(metrics: Dict[str, Tuple[float, float, float]], path: str) -> None:
    metrics.to_csv(path, index=False)


def cross_validate(args, conf, skf, X, y):
    if args.num_classes == 2:
        f1_mean, acc_mean, auc_mean, perc_err_mean, ece_mean = 0, 0, 0, 0, 0
    else:
        micro_mean, macro_mean, weighted_mean, ece_mean = 0, 0, 0, 0
    for train_index, val_index in skf.split(X, y):
        X_train, X_cval = X[train_index], X[val_index]
        y_train, y_cval = y[train_index], y[val_index]
        # Train
        model = train(conf, X_train, y_train, args.val_size, args.seed, args.num_classes)
        # Save metrics
        val_metrics = evaluate(model, X_cval, y_cval, args.num_classes)
        if args.num_classes == 2:
            f1, acc, auc, perc_err, ece = val_metrics
            f1_mean += f1
            acc_mean += acc
            auc_mean += auc
            perc_err_mean += perc_err
            ece_mean += ece
        else:
            micro, macro, weighted, ece = val_metrics
            micro_mean += micro
            macro_mean += macro
            weighted_mean += weighted
            ece_mean += ece
    if args.num_classes == 2:
        f1_mean /= args.cv_folds
        acc_mean /= args.cv_folds
        auc_mean /= args.cv_folds
        perc_err_mean /= args.cv_folds
        ece_mean /= args.cv_folds
        tmp = pd.DataFrame(
            [(*conf.values(), f1_mean, acc_mean, auc_mean, perc_err_mean, ece_mean)],
            columns=[
                'n_estimators', 'learning_rate', 'max_depth', 'colsample_bytree',
                'f1', 'accuracy', 'auc', 'perc_err', 'ece'
                ]
        )
    else:
        micro_mean /= args.cv_folds
        macro_mean /= args.cv_folds
        weighted_mean /= args.cv_folds
        ece_mean /= args.cv_folds
        tmp = pd.DataFrame(
            [(*conf.values(), micro_mean, macro_mean, weighted_mean, ece_mean)],
            columns=[
                'n_estimators', 'learning_rate', 'max_depth', 'colsample_bytree',
                'micro', 'macro', 'weighted', 'ece'
                ]
        )
    return tmp


def create_confs():
    confs = [{'n_estimators': n, 'learning_rate': lr, 'max_depth': d, 'colsample_bytree': c}
             for n in [500]
             for lr in [0.05, 0.005]
             for d in [8, 16]
             for c in [1, 0.5]]
    return confs


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--graph-dir', type=str, required=True,
                        help='Folder containing .graph.csv files.')
    parser.add_argument('--test-graph-dir', type=str, required=True,
                        help='Folder containing .graph.csv files to evaluate at the end.')
    parser.add_argument('--val-size', type=float, default=0.2,
                        help='Validation size used for early stopping. Default: 0.2')
    parser.add_argument('--seed', type=int, default=None,
                        help='Seed for random split. Default: None')
    parser.add_argument('--num-workers', type=int, default=1,
                        help='Number of processors to use. Default: 1.')
    parser.add_argument('--cv-folds', type=int, default=10,
                        help='Number of CV folds. Default: 10.')
    parser.add_argument('--save-name', type=str, required=True,
                        help='Name to save the result, without file type.')
    parser.add_argument('--num-classes', type=int, default=2,
                        help='Number of classes to consider for classification (background not included).')
    return parser


def main_with_args(args: Namespace, logger: Logger):
    X, y = read_all_nodes(args.graph_dir)
    y = np.array(y, dtype=np.int32)
    skf = StratifiedKFold(n_splits=args.cv_folds)
    skf.get_n_splits(X, y)

    if args.num_classes == 2:
        metrics = pd.DataFrame(
            {},
            columns=[
                'n_estimators', 'learning_rate', 'max_depth', 'colsample_bytree',
                'f1', 'accuracy', 'auc', 'perc_err', 'ece'
            ]
        )
    else:
        metrics = pd.DataFrame(
            {},
            columns=[
                'n_estimators', 'learning_rate', 'max_depth', 'colsample_bytree',
                'micro', 'macro', 'weighted', 'ece'
            ]
        )
    confs = create_confs()
    logger.info('Training various XGBoost configurations')
    if args.num_workers > 0:
        with ThreadPoolExecutor(max_workers=args.num_workers) as executor:
            futures = []
            for conf in confs:
                future = executor.submit(cross_validate, args, conf, skf, X, y)
                futures.append(future)
            for k, future in enumerate(futures):
                tmp = future.result()
                logger.info('Configuration {:2}/{:2}'.format(k + 1, len(confs)))
                metrics = pd.concat((metrics, tmp))
                save(metrics, args.save_name + '.csv')
    else:
        for k, conf in enumerate(confs):
            logger.info('Configuration {:2}/{:2}'.format(k + 1, len(confs)))
            tmp = cross_validate(args, conf, skf, X, y)
            metrics = pd.concat((metrics, tmp))
            save(metrics, args.save_name + '.csv')
    save(metrics, args.save_name + '.csv')
    metrics = pd.read_csv(args.save_name + '.csv')
    logger.info('Selecting best XGBoost configuration.')
    if args.num_classes == 2:
        best_conf = metrics.sort_values(by='f1', ascending=False).iloc[0]
    else:
        best_conf = metrics.sort_values(by='weighted', ascending=False).iloc[0]
    tmp = {}
    tmp['n_estimators'] = int(best_conf['n_estimators'])
    tmp['learning_rate'] = float(best_conf['learning_rate'])
    tmp['max_depth'] = int(best_conf['max_depth'])
    tmp['colsample_bytree'] = float(best_conf['colsample_bytree'])
    best_conf = tmp
    logger.info('Retraining with best configuration.')
    model = train(best_conf, X, y, args.val_size, args.seed, args.num_classes)
    logger.info('Computing test metrics.')
    X_test, y_test = read_all_nodes(args.test_graph_dir)
    y_test = np.array(y_test, dtype=np.int32)
    test_metrics = evaluate(model, X_test, y_test, args.num_classes)
    if args.num_classes == 2:
        f1, acc, auc, perc_err, ece = test_metrics
        test_metrics = pd.DataFrame(
            [(*best_conf.values(), f1, acc, auc, perc_err, ece)],
            columns=[
                'n_estimators', 'learning_rate', 'max_depth', 'colsample_bytree',
                'f1', 'accuracy', 'auc', 'perc_err', 'ece'
                ]
        )
    else:
        micro, macro, weighted, ece = test_metrics
        test_metrics = pd.DataFrame(
            [(*best_conf.values(), micro, macro, weighted, ece)],
            columns=[
                'n_estimators', 'learning_rate', 'max_depth', 'colsample_bytree',
                'micro', 'macro', 'weighted', 'ece'
                ]
        )
    save(test_metrics, args.save_name + '_test.csv')


def main():
    parser = _create_parser()
    args = parser.parse_args()

    logger = logging.getLogger('train_xgboost')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    main_with_args(args, logger)



"""
Evaluating script for cell predictions.

Input format
------------
Centroids in a csv with columns X,Y and class. Both for prediction and GT.

Output
------
F1-score (binary, macro and weighted), accuracy, ROC AUC and error percentage between the prediction and the GT at cell-level.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pandas as pd
import numpy as np
from typing import List, Tuple, Dict

from ..utils.preprocessing import read_names, read_centroids
from ..utils.nearest import generate_tree, find_nearest
from ..utils.classification import metrics_from_predictions

import argparse
from argparse import Namespace
import logging
from logging import Logger


def get_confusion_matrix(
        gt_centroids: List[Tuple[int, int, int]],
        pred_centroids: List[Tuple[int, int, int]]
        ) -> np.ndarray:
    """
    Calculates the confusion matrix based on the ground truth and predicted centroids.

    Each centroid is represented by a 3-tuple with (X, Y, class).
    The confusion matrix has (N+1)x(N+1) entries, where N is the maximum value encountered for the class.
    There is one additional entry for the background class when no match is found.

    :param gt_centroids: The ground truth centroids represented as a list of 3-tuples (X, Y, class).
    :type gt_centroids: List[Tuple[int, int, int]]
    :param pred_centroids: The predicted centroids represented as a list of 3-tuples (X, Y, class).
    :type pred_centroids: List[Tuple[int, int, int]]

    :return: The confusion matrix.
    :rtype: np.ndarray
    """
    if len(gt_centroids) == 0 and len(pred_centroids) == 0:
        return np.array([[0]])
    if len(gt_centroids) == 0:
        N = np.max(pred_centroids[:, 2])
        M = np.zeros((N + 1, N + 1))
        classes, freqs = np.unique(pred_centroids[:, 2], return_counts=True)
        M[0, classes] += freqs
        return M
    if len(pred_centroids) == 0:
        N = np.max(gt_centroids[:, 2])
        M = np.zeros((N + 1, N + 1))
        classes, freqs = np.unique(gt_centroids[:, 2], return_counts=True)
        M[classes, 0] += freqs
        return M
    if type(gt_centroids) == list:
        gt_centroids = np.array(gt_centroids)
    if type(pred_centroids) == list:
        pred_centroids = np.array(pred_centroids)
    N = int(max(np.max(gt_centroids[:, 2]), np.max(pred_centroids[:, 2])))
    assert min(np.min(gt_centroids[:, 2]), np.min(pred_centroids[:, 2])) > 0, 'Zero should not be a class.'
    gt_tree = generate_tree(gt_centroids[:, :2])
    pred_tree = generate_tree(pred_centroids[:, :2])
    M = np.zeros((N + 1, N + 1))
    for point_id, point in enumerate(gt_centroids):
        closest_id = find_nearest(point[:2], pred_tree)
        closest = pred_centroids[closest_id]
        if point_id == find_nearest(closest[:2], gt_tree):
            M[int(point[2])][int(closest[2])] += 1  # 1-1 matchings
        else:
            M[int(point[2])][0] += 1  # GT not matched in prediction
    for point_id, point in enumerate(pred_centroids):
        closest_id = find_nearest(point[:2], gt_tree)
        closest = gt_centroids[closest_id]
        if point_id != find_nearest(closest[:2], pred_tree):
            M[0][int(point[2])] += 1  # Prediction not matched in GT
    return M


def get_pairs(
        gt_centroids: List[Tuple[int, int, int]],
        pred_centroids: List[Tuple[int, int, int]]
        ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Retrieves true and predicted labels ordered by their correspondences.

    Each centroid is represented by a 3-tuple with (X, Y, class).
    The class is represented as 1=non-tumour, 2=tumour.
    The returned labels are ordered starting from 0.

    :param gt_centroids: The ground truth centroids represented as a list of 3-tuples (X, Y, class).
    :type gt_centroids: List[Tuple[int, int, int]]
    :param pred_centroids: The predicted centroids represented as a list of 3-tuples (X, Y, class).
    :type pred_centroids: List[Tuple[int, int, int]]

    :return: A tuple containing the true labels and predicted labels ordered by their correspondences.
    :rtype: Tuple[np.ndarray, np.ndarray]
    """
    if len(gt_centroids) == 0:
        return None, None

    gt_tree = generate_tree(gt_centroids[:, :2])
    pred_tree = generate_tree(pred_centroids[:, :2])
    true_labels, pred_labels = [], []
    for point_id, point in enumerate(gt_centroids):
        closest_id = find_nearest(point[:2], pred_tree)
        closest = pred_centroids[closest_id]
        if closest[2] != -1 and point[2] != -1 and point_id == find_nearest(closest[:2], gt_tree):
            true_labels.append(point[2] - 1)
            pred_labels.append(closest[2] - 1)
    return np.array(true_labels), np.array(pred_labels)


def compute_f1_score_from_matrix(conf_mat: np.ndarray, cls: int) -> float:
    """
    Computes the F1 score of a given class against the rest.

    :param conf_mat: The confusion matrix.
    :type conf_mat: np.ndarray
    :param cls: The class for which to compute the F1 score.
    :type cls: int

    :return: The F1 score of the given class against the rest.
    :rtype: float
    """
    if cls == 0:
        return None
    TP = conf_mat[cls, cls]
    PP = conf_mat[:, cls].sum()
    if PP == 0:
        return None
    P = conf_mat[cls, :].sum()
    if P == 0:
        return None
    precision = TP / PP
    recall = TP / P
    if TP == 0:
        return 0
    return 2 * precision * recall / (precision + recall)


def compute_metrics_from_matrix(conf_mat: np.ndarray) -> Tuple[float, float, float, float]:
    """
    Computes various evaluation metrics from a confusion matrix.

    Given a confusion matrix, this function calculates the F1 score, accuracy,
    ROC AUC, and percentage error.

    :param conf_mat: The confusion matrix.
    :type conf_mat: np.ndarray

    :return: A tuple containing the F1 score, accuracy, ROC AUC, and percentage error.
    :rtype: Tuple[float, float, float, float]
    """
    total = np.sum(conf_mat)
    acc = np.matrix.trace(conf_mat) / total
    n_classes = len(conf_mat)
    macro, weighted = 0, 0
    real_n_classes = n_classes
    adjust_weighted = 1
    for k in range(n_classes):
        f1_class = compute_f1_score_from_matrix(conf_mat, cls=k)
        cls_supp = conf_mat[k, :].sum() / total
        if f1_class is not None:
            macro += f1_class
            weighted += f1_class * cls_supp
        else:
            real_n_classes -= 1
            adjust_weighted -= cls_supp
    macro /= real_n_classes
    weighted /= adjust_weighted
    return macro, weighted, acc


def add_matrices(A: np.ndarray, B: np.ndarray) -> np.ndarray:
    """
    Adds two matrices even if their shapes are different.
    If B is bigger than A, A is enlarged with zeros.
    And vice versa.

    :param A: The first matrix.
    :type A: np.ndarray
    :param B: The second matrix.
    :type B: np.ndarray

    :return: The sum of the two matrices.
    :rtype: np.ndarray
    """
    if A.shape == B.shape:
        return A + B
    n, m = A.shape[0], B.shape[0]
    if m > n:
        res = np.zeros((m, m))
        res[:n, :n] += A
        res += B
    else:
        res = np.zeros((n, n))
        res[:m, :m] += B
        res += A
    return res


def compute_perc_error_bkgr(
        gt_centroids: List[Tuple[int, int, int]],
        pred_centroids: List[Tuple[int, int, int]]
        ) -> float:
    """
    Computes absolute difference in percentages of the classes.

    :param gt_centroids: Ground truth.
    :type gt_centroids: List[Tuple[int, int, int]]
    :param pred_centroids: Predictions.
    :type pred_centroids: List[Tuple[int, int, int]]
    :return: The percentage error.
    :rtype: float
    """
    gt_cls = gt_centroids[:, 2]
    pred_cls = pred_centroids[:, 2]
    gt_perc = (gt_cls == 2).sum() / len(gt_cls)
    pred_perc = (pred_cls == 2).sum() / len(pred_cls)
    return abs(gt_perc - pred_perc)


def save_csv(
        metrics: Dict[str, List[float]],
        save_path: str
        ) -> None:
    """
    Saves metrics in CSV format for later use.

    :param metrics: The metrics dictionary containing the data to be saved.
    :type metrics: Dict[str, List[float]]
    :param save_path: The file path where the CSV file will be saved.
    :type save_path: str
    """
    metrics_df = pd.DataFrame(metrics)
    metrics_df.to_csv(save_path + '.csv', index=False)


def save_debug_matrix(
        mat: np.ndarray,
        save_path: str
        ) -> None:
    """
    Saves a confusion matrix in CSV format for debug purposes.

    :param mat: The confusion matrix to be saved.
    :type mat: np.ndarray
    :param save_path: The file path where the CSV file will be saved.
    :type save_path: str
    """
    conf_mat_df = pd.DataFrame(mat)
    conf_mat_df.to_csv(save_path + '.csv')


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--names', type=str, required=True,
                        help='Path to txt file with names.')
    parser.add_argument('--gt-path', type=str, required=True,
                        help='Path to GT files.')
    parser.add_argument('--pred-path', type=str, required=True,
                        help='Path to prediction files.')
    parser.add_argument('--save-name', type=str, required=True,
                        help='Name to save the result, without file type.')
    parser.add_argument('--debug-path', type=str, default=None,
                        help='Name of file where to save confusion matrices (optional).')
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    return parser


def main_with_args(args: Namespace, logger: Logger):
    names = read_names(args.names)
    if args.num_classes == 2:
        metrics = {
            'Name': [], 'F1': [], 'Accuracy': [], 'ROC_AUC': [], 'Perc_err': [], 'ECE': [],
            'Macro F1 (bkgr)': [], 'Weighted F1 (bkgr)': [], 'Micro F1 (bkgr)': [], 'Perc_err (bkgr)': []
        }
    else:
        metrics = {
            'Name': [], 'Macro F1': [], 'Weighted F1': [], 'Micro F1': [], 'ECE': [],
            'Macro F1 (bkgr)': [], 'Weighted F1 (bkgr)': [], 'Micro F1 (bkgr)': []
        }
    global_conf_mat = None
    global_pred, global_true = [], []
    global_gt_centroids, global_pred_centroids = np.array([]).reshape(0,3), np.array([]).reshape(0,3)
    for k, name in enumerate(names):
        logger.info('Progress: {:2d}/{}'.format(k + 1, len(names)))
        metrics['Name'].append(name)
        # Read
        gt_centroids = read_centroids(name, args.gt_path)
        pred_centroids = read_centroids(name, args.pred_path)
        global_gt_centroids = np.vstack((global_gt_centroids, gt_centroids))
        global_pred_centroids = np.vstack((global_pred_centroids, pred_centroids))
        # Compute pairs and confusion matrix
        conf_mat = get_confusion_matrix(gt_centroids, pred_centroids)
        if args.debug_path is not None:
            save_debug_matrix(conf_mat, args.debug_path + '_' + name)
        try:
            true_labels, pred_labels = get_pairs(gt_centroids, pred_centroids)
        except Exception as e:
            logger.error(e)
            logger.error(name)
            _metrics = [-1] * 5 if args.num_classes == 2 else [-1] * 4
            macro_bkgr, weighted_bkgr, micro_bkgr = -1, -1, -1
            true_labels, pred_labels = None, None
        # Save for later
        if true_labels is not None and pred_labels is not None:
            global_true.extend(true_labels)
            global_pred.extend(pred_labels)
        if global_conf_mat is None:
            global_conf_mat = conf_mat
        else:
            global_conf_mat = add_matrices(global_conf_mat, conf_mat)
        # Compute per image scores and percentages
        try:
            _metrics = metrics_from_predictions(true_labels, pred_labels, None, args.num_classes)
            macro_bkgr, weighted_bkgr, micro_bkgr = compute_metrics_from_matrix(conf_mat)
            if args.num_classes == 2:
                perc_error_bkgr = compute_perc_error_bkgr(gt_centroids, pred_centroids)
        except Exception as e:
            logger.error(e)
            logger.error(name)
            _metrics = [-1] * 5 if args.num_classes == 2 else [-1] * 4
            macro_bkgr, weighted_bkgr, micro_bkgr = -1, -1, -1
        if args.num_classes == 2:
            acc, f1, auc, perc_error, ece = _metrics
            metrics['F1'].append(f1)
            metrics['Accuracy'].append(acc)
            metrics['ROC_AUC'].append(auc)
            metrics['Perc_err'].append(perc_error)
            metrics['ECE'].append(ece)
            metrics['Macro F1 (bkgr)'].append(macro_bkgr)
            metrics['Weighted F1 (bkgr)'].append(weighted_bkgr)
            metrics['Micro F1 (bkgr)'].append(micro_bkgr)
            metrics['Perc_err (bkgr)'].append(perc_error_bkgr)
        else:
            micro, macro, weighted, ece = _metrics
            metrics['Macro F1'].append(macro)
            metrics['Weighted F1'].append(weighted)
            metrics['Micro F1'].append(micro)
            metrics['ECE'].append(ece)
            metrics['Macro F1 (bkgr)'].append(macro_bkgr)
            metrics['Weighted F1 (bkgr)'].append(weighted_bkgr)
            metrics['Micro F1 (bkgr)'].append(micro_bkgr)
    if args.debug_path is not None:
        save_debug_matrix(global_conf_mat, args.debug_path + '_global')
    save_csv(metrics, args.save_name)
    # Global scores and percentages
    _global_metrics = metrics_from_predictions(np.array(global_true), np.array(global_pred), None, args.num_classes)
    macro_bkgr, weighted_bkgr, micro_bkgr = compute_metrics_from_matrix(global_conf_mat)
    if args.num_classes == 2:
        acc, f1, auc, perc_error, ece = _global_metrics
        perc_error_bkgr = compute_perc_error_bkgr(global_gt_centroids, global_pred_centroids)
        global_metrics = {
            'Name': ['All'], 'F1': [f1], 'Accuracy': [acc], 'ROC_AUC': [auc], 'Perc_err': [perc_error], 'ECE': [ece],
            'Macro F1 (bkgr)': [macro_bkgr], 'Weighted F1 (bkgr)': [weighted_bkgr], 'Micro F1 (bkgr)': [micro_bkgr],
            'Perc_err (bkgr)': [perc_error_bkgr]
        }
    else:
        micro, macro, weighted, ece = _global_metrics
        global_metrics = {
            'Name': ['All'], 'Macro F1': [macro], 'Weighted F1': [weighted], 'Micro F1': [micro], 'ECE': [ece],
            'Macro F1 (bkgr)': [macro_bkgr], 'Weighted F1 (bkgr)': [weighted_bkgr], 'Micro F1 (bkgr)': [micro_bkgr]
        }
    save_csv(global_metrics, args.save_name + '_all')


def main():
    parser = _create_parser()
    args = parser.parse_args()

    logger = logging.getLogger('eval_segment')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    main(args, logger)



"""
Graph Representation Learning via Hard Attention Networks in DGL using Adam optimization,
with some minor modifications.

References
----------
Paper: https://arxiv.org/abs/1907.04652

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""

from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F

import dgl.function as fn
from dgl.base import DGLError
from dgl.nn.pytorch import edge_softmax
from dgl.nn.pytorch.utils import Identity
from dgl.sampling import select_topk
from .norm import Norm


class HardGAO(nn.Module):
    def __init__(
        self,
        in_feats,
        out_feats,
        num_heads=8,
        feat_drop=0.0,
        attn_drop=0.0,
        negative_slope=0.2,
        residual=True,
        activation=F.elu,
        k=8,
    ):
        super(HardGAO, self).__init__()
        self.num_heads = num_heads
        self.in_feats = in_feats
        self.out_feats = out_feats
        self.k = k
        self.residual = residual
        # Initialize Parameters for Additive Attention
        self.fc = nn.Linear(
            self.in_feats, self.out_feats * self.num_heads, bias=False
        )
        self.attn_l = nn.Parameter(
            torch.FloatTensor(size=(1, self.num_heads, self.out_feats))
        )
        self.attn_r = nn.Parameter(
            torch.FloatTensor(size=(1, self.num_heads, self.out_feats))
        )
        # Initialize Parameters for Hard Projection
        self.p = nn.Parameter(torch.FloatTensor(size=(1, in_feats)))
        # Initialize Dropouts
        self.feat_drop = nn.Dropout(feat_drop)
        self.attn_drop = nn.Dropout(attn_drop)
        self.leaky_relu = nn.LeakyReLU(negative_slope)
        if self.residual:
            if self.in_feats == self.out_feats:
                self.residual_module = Identity()
            else:
                self.residual_module = nn.Linear(
                    self.in_feats, self.out_feats * num_heads, bias=False
                )

        self.reset_parameters()
        self.activation = activation

    def reset_parameters(self):
        gain = nn.init.calculate_gain("relu")
        nn.init.xavier_normal_(self.fc.weight, gain=gain)
        nn.init.xavier_normal_(self.p, gain=gain)
        nn.init.xavier_normal_(self.attn_l, gain=gain)
        nn.init.xavier_normal_(self.attn_r, gain=gain)
        if self.residual:
            nn.init.xavier_normal_(self.residual_module.weight, gain=gain)

    def forward(self, graph, feat, get_attention=False):
        # Check in degree and generate error
        if (graph.in_degrees() == 0).any():
            raise DGLError(
                "There are 0-in-degree nodes in the graph, "
                "output for those nodes will be invalid. "
                "This is harmful for some applications, "
                "causing silent performance regression. "
                "Adding self-loop on the input graph by "
                "calling `g = dgl.add_self_loop(g)` will resolve "
                "the issue. Setting ``allow_zero_in_degree`` "
                "to be `True` when constructing this module will "
                "suppress the check and let the code run."
            )
        # projection process to get importance vector y
        graph.ndata["y_"] = torch.abs(
            torch.matmul(self.p, feat.T).view(-1)
        ) / torch.norm(self.p, p=2)
        # Use edge message passing function to get the weight from src node
        graph.apply_edges(fn.copy_u("y_", "y_"))
        # Select Top k neighbors
        subgraph = select_topk(graph.cpu(), self.k, "y_").to(graph.device)
        # Sigmoid as information threshold
        subgraph.ndata["y_"] = torch.sigmoid(subgraph.ndata["y_"])
        # Using vector matrix elementwise mul for acceleration
        feat = subgraph.ndata["y_"].view(-1, 1) * feat
        feat = self.feat_drop(feat)
        h = self.fc(feat).view(-1, self.num_heads, self.out_feats)
        el = (h * self.attn_l).sum(dim=-1).unsqueeze(-1)
        er = (h * self.attn_r).sum(dim=-1).unsqueeze(-1)
        # Assign the value on the subgraph
        subgraph.srcdata.update({"ft": h, "el": el})
        subgraph.dstdata.update({"er": er})
        # compute edge attention, el and er are a_l Wh_i and a_r Wh_j respectively.
        subgraph.apply_edges(fn.u_add_v("el", "er", "e"))
        e = self.leaky_relu(subgraph.edata.pop("e"))
        # compute softmax
        subgraph.edata["a"] = self.attn_drop(edge_softmax(subgraph, e))
        # message passing
        subgraph.update_all(fn.u_mul_e("ft", "a", "m"), fn.sum("m", "ft"))
        rst = subgraph.dstdata["ft"]
        # activation
        if self.activation:
            rst = self.activation(rst)
        # Residual
        if self.residual:
            rst = rst + self.residual_module(feat).view(
                feat.shape[0], -1, self.out_feats
            )

        if get_attention:
            return rst, subgraph.edata["a"]
        else:
            return rst


class HardGAT(nn.Module):
    def __init__(
        self,
        in_dim,
        num_hidden,
        num_classes,
        heads,
        num_layers,
        dropout,
        norm_type,
        activation=F.elu,
        negative_slope=0.2,
        residual=False,
        k=8,
        enable_background=False,
    ):
        super(HardGAT, self).__init__()
        feat_drop=dropout
        attn_drop=dropout
        self.num_layers = num_layers
        self.gat_layers = nn.ModuleList()
        self.activation = activation
        gat_layer = partial(HardGAO, k=k)
        # input projection (no residual)
        self.gat_layers.append(
            gat_layer(
                in_dim,
                num_hidden,
                heads[0],
                feat_drop,
                attn_drop,
                negative_slope,
                False,
                self.activation,
            )
        )
        self.gat_layers.append(Norm(norm_type=norm_type, hidden_dim=num_hidden * heads[0]))
        # hidden layers
        for l in range(1, num_layers):
            # due to multi-head, the in_dim = num_hidden * num_heads
            self.gat_layers.append(
                gat_layer(
                    num_hidden * heads[l - 1],
                    num_hidden,
                    heads[l],
                    feat_drop,
                    attn_drop,
                    negative_slope,
                    residual,
                    self.activation,
                )
            )
            self.gat_layers.append(Norm(norm_type=norm_type, hidden_dim=num_hidden * heads[l]))
        # output projection
        self.gat_layers.append(
            gat_layer(
                num_hidden * heads[-2],
                num_classes,
                heads[-1],
                feat_drop,
                attn_drop,
                negative_slope,
                False,
                None,
            )
        )

        self.enable_background = enable_background
        if enable_background:
            self.bkgr_head = gat_layer(
                num_hidden * heads[-2],
                2,
                heads[-1],
                feat_drop,
                attn_drop,
                negative_slope,
                False,
                None,
            )

    def forward(self, g, inputs):
        h = inputs
        for l in range(self.num_layers):
            h = self.gat_layers[l](g, h).flatten(1)
        logits = self.gat_layers[-1](g, h).mean(1)
        if self.enable_background:
            logits_bkgr = self.bkgr_head(g, h).mean(1)
            return logits, logits_bkgr
        return logits



import os
import random
import shutil
from collections import OrderedDict
import numpy as np
import torch
from termcolor import colored


####
def convert_pytorch_checkpoint(net_state_dict):
    variable_name_list = list(net_state_dict.keys())
    is_in_parallel_mode = all(v.split(".")[0] == "module" for v in variable_name_list)
    if is_in_parallel_mode:
        colored_word = colored("WARNING", color="red", attrs=["bold"])
        print(
            (
                "%s: Detect checkpoint saved in data-parallel mode."
                " Converting saved model to single GPU mode." % colored_word
            ).rjust(80)
        )
        net_state_dict = {
            ".".join(k.split(".")[1:]): v for k, v in net_state_dict.items()
        }
    return net_state_dict


####
def check_manual_seed(seed):
    """ If manual seed is not specified, choose a 
    random one and communicate it to the user.

    Args:
        seed: seed to check

    """
    seed = seed or random.randint(1, 10000)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    # ia.random.seed(seed)

    print("Using manual seed: {seed}".format(seed=seed))
    return


####
def check_log_dir(log_dir):
    """Check if log directory exists.

    Args:
        log_dir: path to logs

    """
    if os.path.isdir(log_dir):
        colored_word = colored("WARNING", color="red", attrs=["bold", "blink"])
        print("%s: %s exist!" % (colored_word, colored(log_dir, attrs=["underline"])))
        while True:
            print("Select Action: d (delete) / q (quit)", end="")
            key = input()
            if key == "d":
                shutil.rmtree(log_dir)
                break
            elif key == "q":
                exit()
            else:
                color_word = colored("ERR", color="red")
                print("---[%s] Unrecognize Characters!" % colored_word)
    return


def get_model_summary(
    model, input_size, batch_size=-1, device=torch.device("cpu"), dtypes=None
):
    """Reusable utility layers such as pool or upsample will also get printed, but their printed values will
    be corresponding to the last call.

    """
    if dtypes == None:
        dtypes = [torch.FloatTensor] * len(input_size)

    summary_str = ""

    def register_hook(module):
        def hook(module, input, output):
            class_name = str(module.__class__).split(".")[-1].split("'")[0]
            module_idx = len(summary)

            m_key = module.name if module.name != "" else "%s" % class_name

            summary[m_key] = OrderedDict()
            summary[m_key]["input_shape"] = list(input[0].size())
            summary[m_key]["input_shape"][0] = batch_size
            if isinstance(output, (list, tuple)):
                summary[m_key]["output_shape"] = [
                    [-1] + list(o.size())[1:] for o in output
                ]
            elif isinstance(output, dict):
                summary[m_key]["output_shape"] = [
                    [-1] + list(o.size())[1:] for o in output.values()
                ]
            elif isinstance(output, torch.Tensor):
                summary[m_key]["output_shape"] = list(output.size())
                summary[m_key]["output_shape"][0] = batch_size

            params = 0
            if hasattr(module, "weight") and hasattr(module.weight, "size"):
                params += torch.prod(torch.LongTensor(list(module.weight.size())))
                summary[m_key]["trainable"] = module.weight.requires_grad
            if hasattr(module, "bias") and hasattr(module.bias, "size"):
                params += torch.prod(torch.LongTensor(list(module.bias.size())))
            summary[m_key]["nb_params"] = params

        if len(list(module.children())) == 0:
            hooks.append(module.register_forward_hook(hook))

    # multiple inputs to the network
    if isinstance(input_size, tuple):
        input_size = [input_size]

    # batch_size of 2 for batchnorm
    x = [
        torch.rand(2, *in_size).type(dtype).to(device=device)
        for in_size, dtype in zip(input_size, dtypes)
    ]

    # create properties
    summary = OrderedDict()
    hooks = []

    # create layer name according to hierachy names
    for name, module in model.named_modules():
        module.name = name

    # register hook
    model.apply(register_hook)

    # make a forward pass
    model(*x)

    # aligning name to the left
    max_name_length = len(max(summary.keys(), key=len))
    summary = [(k.ljust(max_name_length), v) for k, v in summary.items()]
    summary = OrderedDict(summary)

    # remove these hooks
    for h in hooks:
        h.remove()

    header_line = "{}  {:>25} {:>15}".format(
        "Layer Name".center(max_name_length), "Output Shape", "Param #"
    )
    summary_str += "".join("-" for _ in range(len(header_line))) + "\n"
    summary_str += header_line + "\n"
    summary_str += "".join("=" for _ in range(len(header_line))) + "\n"
    total_params = 0
    total_output = 0
    trainable_params = 0
    for layer in summary:
        # input_shape, output_shape, trainable, nb_params
        line_new = "{:>20}  {:>25} {:>15}".format(
            layer,
            str(summary[layer]["output_shape"]),
            "{0:,}".format(summary[layer]["nb_params"]),
        )
        total_params += summary[layer]["nb_params"]

        total_output += np.prod(summary[layer]["output_shape"])
        if "trainable" in summary[layer]:
            if summary[layer]["trainable"] == True:
                trainable_params += summary[layer]["nb_params"]
        summary_str += line_new + "\n"

    # assume 4 bytes/number (float on cuda).
    total_input_size = abs(
        np.prod(sum(input_size, ())) * batch_size * 4.0 / (1024 ** 2.0)
    )
    total_output_size = abs(
        2.0 * total_output * 4.0 / (1024 ** 2.0)
    )  # x2 for gradients
    total_params_size = abs(total_params * 4.0 / (1024 ** 2.0))
    total_size = total_params_size + total_output_size + total_input_size

    summary_str += "".join("=" for _ in range(len(header_line))) + "\n"
    summary_str += "Total params: {0:,}".format(total_params) + "\n"
    summary_str += "Trainable params: {0:,}".format(trainable_params) + "\n"
    summary_str += (
        "Non-trainable params: {0:,}".format(total_params - trainable_params) + "\n"
    )
    summary_str += "".join("-" for _ in range(len(header_line))) + "\n"
    summary_str += "Input size (MB): %0.2f" % total_input_size + "\n"
    summary_str += "Forward/backward pass size (MB): %0.2f" % total_output_size + "\n"
    summary_str += "Params size (MB): %0.2f" % total_params_size + "\n"
    summary_str += "Estimated Total Size (MB): %0.2f" % total_size + "\n"
    summary_str += "".join("-" for _ in range(len(header_line))) + "\n"
    return summary_str



"""
Module with utility functions for the classification module.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import numpy as np
from sklearn.preprocessing import StandardScaler
from typing import Tuple, List, Optional
from sklearn.metrics import roc_auc_score, f1_score, accuracy_score
from .read_nodes import _read_all_nodes
from .calibration_error import calibration_error


def normalize(X_train: np.ndarray, X_val: np.ndarray, X_test: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Normalizes the input data using the mean and standard deviation of the training dataset.

    :param X_train: The input training data with shape (n_samples, n_features).
    :type X_train: numpy.ndarray
    :param X_val: The input validation data with shape (n_samples, n_features).
    :type X_val: numpy.ndarray
    :param X_test: The input test data with shape (n_samples, n_features).
    :type X_test: numpy.ndarray
    :return: A tuple containing three numpy.ndarrays:
             - The normalized X_train with shape (n_samples, n_features).
             - The normalized X_val with shape (n_samples, n_features).
             - The normalized X_test with shape (n_samples, n_features).
    :rtype: tuple(numpy.ndarray, numpy.ndarray, numpy.ndarray)
    """
    sc = StandardScaler()
    sc.fit(X_train)
    return sc.transform(X_train), sc.transform(X_val), sc.transform(X_test)


def fit_column_normalizer(
        node_dir: str,
        names: List[str],
        remove_prior: Optional[bool] = False,
        remove_morph: Optional[bool] = False,
        enable_background: Optional[bool] = False,
        ) -> StandardScaler:
    """
    Fits a StandardScaler object to the input data and returns it.

    :param node_dir: The directory containing node data files.
    :type node_dir: str
    :param names: A list of node names to read from the node directory.
    :type names: List[str]
    :param remove_prior: If True, hovernet probabilites are ignored.
    :type remove_prior: Optional[bool]
    :param remove_morph: If True, morphological features are removed.
    :type remove_morph: Optional[bool]
    :return: A StandardScaler object fitted to the input data.
    :rtype: StandardScaler
    """
    X, y = _read_all_nodes(node_dir, [x + '.nodes.csv' for x in names], remove_morph=remove_morph, remove_prior=remove_prior, enable_background=enable_background)
    sc = StandardScaler()
    sc.fit(X)
    return sc


def check_imbalance(y_true: np.ndarray) -> bool:
    """
    Returns true if there are at least two classes, and false otherwise.
    """
    return len(np.unique(y_true)) > 1


def percentage_error(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Computes the deviation in the percentage of tumoral cells.
    """
    gt_perc = (y_true == 1).sum() / len(y_true)
    pred_perc = (y_pred == 1).sum() / len(y_pred)
    return abs(gt_perc - pred_perc)


def metrics_from_predictions(
        y_true: np.ndarray,
        y_pred: np.ndarray,
        y_prob: Optional[np.ndarray] = None,
        num_classes: Optional[int] = 2,
        ) -> List[float]:
    """
    Computes evaluation metrics using predictions and ground truth.
    For binary: Accuracy, F1 Score, ROC AUC, %ERR, ECE
    For multiclass: Micro F1, Macro F1, Weighted F1, ECE

    :param y_true: The ground truth to compare with, values start at 0. Shape: (N, 1).
    :type y_true: np.ndarray
    :param y_pred: Predictions with values starting at 0. Shape: (N, 1).
    :type y_pred: np.ndarray
    :param y_prob: Probabilities of the predictions. Shape: (N, num_classes) or (N, 1) if binary.
    :type y_prob: Optional[np.ndarray]
    :param num_classes: Number of classes to consider.
    :type num_classes: Optional[int]
    :return: Several metrics. Either 4 or 5 floats, depending on whether is binary or multiclass.
    :rtype: List[float]
    """
    if num_classes == 2:
        acc = accuracy_score(y_true, y_pred)
        f1 = f1_score(y_true, y_pred, zero_division=0)
        if check_imbalance(y_true):
            if y_prob is not None:
                auc = roc_auc_score(y_true, y_prob)
            else:
                auc = roc_auc_score(y_true, y_pred)
        else:
            auc = -1
        perc_error = percentage_error(y_true, y_pred)
        if y_prob is not None:
            ece = calibration_error(y_true, y_prob, norm='l1', reduce_bias=False)
        else:
            ece = -1
        return acc, f1, auc, perc_error, ece
    else:
        micro = f1_score(y_true, y_pred, average='micro')
        macro = f1_score(y_true, y_pred, average='macro')
        weighted = f1_score(y_true, y_pred, average='weighted')
        if y_prob is not None:
            ece = 0
            for i in range(num_classes):
                y_true_i = (y_true == i) * 1
                y_prob_i = y_prob[:, i]
                ece_i = calibration_error(y_true_i, y_prob_i, norm='l1', reduce_bias=False)
                ece += ece_i
            ece /= num_classes
        else:
            ece = -1
        return micro, macro, weighted, ece



import argparse
import pandas as pd
import numpy as np


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--file')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()
    mat = pd.read_csv(args.file, index_col=0).to_numpy()
    total = mat.sum()
    pred = mat[:, 2].sum()
    true = mat[2, :].sum()
    pred_perc = pred / total * 100
    true_perc = true / total * 100
    print(f'True percentage: {true_perc:.2f}\%.')
    print(f'Predicted percentage: {pred_perc:.2f}\%.')
    print(f'Difference: {np.abs(true_perc - pred_perc):.2f}\%.')


if __name__ == '__main__':
    main()


"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import numpy as np
import cProfile
import io
import pstats
from ..utils.postprocessing import get_N_closest_pairs_dists


THRESHOLD = 0.001
tests = [
    (
        np.hstack((np.linspace(0, 30, 100).reshape(-1, 1), np.ones((100, 1)))),
        np.hstack((np.linspace(0, 30, 100).reshape(-1, 1), np.ones((100, 1)) * 2)),
        np.ones((10,))
    ),
    (
        np.hstack((np.linspace(0, 300, 101).reshape(-1, 1), np.ones((101, 1)))) + np.array([[1, 0]]),
        np.hstack((np.linspace(0, 300, 101).reshape(-1, 1), np.ones((101, 1)) * 2)),
        np.ones((20,)) * np.sqrt(2)
    ),
    (
        np.hstack((np.linspace(0, 30, 11).reshape(-1, 1)**2, np.ones((11, 1)))),
        np.hstack((np.linspace(0, 30, 11).reshape(-1, 1), np.ones((11, 1)) * 2)),
        np.array([1, 1, np.sqrt(3**2 + 1), np.sqrt(3**2 + 1), np.sqrt(3**2 + 1), np.sqrt(6**2 + 1), np.sqrt(6**2 + 1)])
    ),
    (
        np.hstack((np.linspace(0, 300000, 100001).reshape(-1, 1), np.ones((100001, 1)))) + np.array([[1, 0]]),
        np.hstack((np.linspace(0, 300, 101).reshape(-1, 1), np.ones((101, 1)) * 2)),
        np.ones((20,)) * np.sqrt(2)
    )
]


def exec_conversion(A, B, result):
    pred = get_N_closest_pairs_dists(A, B, len(result))
    pred = sorted(pred)
    if len(pred) != len(result):
        return False
    for i in range(len(pred)):
        if abs(pred[i] - result[i]) > THRESHOLD:
            return False
    return True

# cProfile.runctx('exec_conversion(*tests[0])', None, locals(), filename='prof1.txt')
# cProfile.runctx('exec_conversion(*tests[1])', None, locals(), filename='prof2.txt')
# cProfile.runctx('exec_conversion(*tests[2])', None, locals(), filename='prof3.txt')


for i in range(len(tests)):
    pr = cProfile.Profile()
    pr.enable()

    my_result = exec_conversion(*tests[i])

    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats('tottime')
    ps.print_stats()

    with open('cpairs/cpairs_profile' + str(i + 1) + '.txt', 'w+') as f:
        f.write(s.getvalue())



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import Callable, Tuple, List


Point = Tuple[float, float]
Contour = List[Point]
Cell = Tuple[int, int, Contour]  # id, class


def rswoosh(
        I: List[Cell],
        isEqual: Callable[[Contour, Contour], bool],
        merge: Callable[[Cell, Cell], Cell]
        ) -> List[Cell]:
    """
    Given a list of cells, returns the same list without duplicates.

    The `rswoosh` function takes a list of cells `I`, a function `isEqual` to check equality between contours,
    and a function `merge` to merge two cells.

    :param I: The list of cells.
    :type I: List[Cell]
    :param isEqual: A function to check equality between contours.
    :type isEqual: Callable[[Contour, Contour], bool]
    :param merge: A function to merge two cells.
    :type merge: Callable[[Cell, Cell], Cell]

    :return: The list of cells without duplicates.
    :rtype: List[Cell]
    """
    out = {}  # Hash Table
    while (len(I) > 0):
        key_r, class_r, r = I.pop()  # O(1)
        exist_equal = False
        for key_s, (class_s, s) in out.items():
            if isEqual(r, s):
                del out[key_s]  # O(1)
                I.append(merge((key_r, class_r, r), (key_s, class_s, s)))  # O(1)
                exist_equal = True
                break
        if not exist_equal:
            out[key_r] = (class_r, r)  # O(1)
    return out.items()



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import numpy as np
import os
import json
import random

TEST_DIR = os.path.dirname(os.path.abspath(__file__))

def save(A, B, name):
    with open(TEST_DIR + '/rswoosh/' + name + '.input.json', 'w') as f:
        json.dump(A, f)
    with open(TEST_DIR + '/rswoosh/' + name + '.output.json', 'w') as f:
        json.dump(B, f)

def create_line(p, q):
    x = np.linspace(p[0], q[0], 100)[1:-1]
    y = np.linspace(p[1], q[1], 100)[1:-1]
    return np.hstack((x.reshape(-1,1), y.reshape(-1,1)))

def split_circle(radius, theta):
    tol = 1e-5
    left_angles = np.linspace(theta + tol, theta + np.pi, 100)
    left_points = np.hstack((np.cos(left_angles).reshape(-1,1), np.sin(left_angles).reshape(-1,1))) * radius
    right_angles = np.linspace(theta + np.pi + tol, theta + 2*np.pi, 100)
    right_points = np.hstack((np.cos(right_angles).reshape(-1,1), np.sin(right_angles).reshape(-1,1))) * radius
    merged = np.vstack((left_points, right_points)).copy()
    left_points = np.vstack((left_points, create_line(left_points[-1], left_points[0])))
    right_points = np.vstack((right_points, create_line(right_points[-1], right_points[0])))
    return left_points, right_points, merged

def translate(points, x, y):
    points = points.copy()
    points[:,0] += x
    points[:,1] += y
    return points.tolist()

def generate_cell_tuple(points, x, y):
    idx = 1e5 * x + y
    cls = 0
    return tuple((int(idx),int(cls),points))

if __name__=='__main__':
    """
    First test:
    Circles split in two with a line in between.
    """
    counter = 0
    for r in [10,20,30]:
        for theta in [0, np.pi / 8, np.pi / 4]:
            counter += 1
            left, right, merged = split_circle(r, theta)
            A, B = [], []
            for x in range(0,16*r,4*r):
                for y in range(0,16*r,4*r):
                    rand_number = random.randint(0,5)
                    if rand_number == 3:
                        A.append(generate_cell_tuple(translate(left, x, y), x / r, y / r))
                        A.append(generate_cell_tuple(translate(right, x, y), x / r, y / r + 1))
                        B.append(generate_cell_tuple(translate(merged, x, y), x / r, y / r))
                    elif rand_number == 1:
                        A.append(generate_cell_tuple(translate(left, x, y), x / r, y / r))
                        B.append(generate_cell_tuple(translate(left, x, y), x / r, y / r))
                    else:
                        A.append(generate_cell_tuple(translate(merged, x, y), x / r, y / r))
                        B.append(generate_cell_tuple(translate(merged, x, y), x / r, y / r))
            save(A,B, 'circles'+str(counter))


from collections import OrderedDict
import torch
import torch.nn as nn

from .net_utils import (DenseBlock, Net, ResidualBlock, TFSamepaddingLayer,
                        UpSample2x)
from .utils import crop_op

####
class HoVerNet(Net):
    """Initialise HoVer-Net."""

    def __init__(self, input_ch=3, nr_types=None, freeze=False, mode='original'):
        super().__init__()
        self.mode = mode
        self.freeze = freeze
        self.nr_types = nr_types
        self.output_ch = 3 if nr_types is None else 4

        assert mode == 'original' or mode == 'fast', \
                'Unknown mode `%s` for HoVerNet %s. Only support `original` or `fast`.' % mode

        module_list = [
            ("/", nn.Conv2d(input_ch, 64, 7, stride=1, padding=0, bias=False)),
            ("bn", nn.BatchNorm2d(64, eps=1e-5)),
            ("relu", nn.ReLU(inplace=True)),
        ]
        if mode == 'fast': # prepend the padding for `fast` mode
            module_list = [("pad", TFSamepaddingLayer(ksize=7, stride=1))] + module_list

        self.conv0 = nn.Sequential(OrderedDict(module_list))
        self.d0 = ResidualBlock(64, [1, 3, 1], [64, 64, 256], 3, stride=1)
        self.d1 = ResidualBlock(256, [1, 3, 1], [128, 128, 512], 4, stride=2)
        self.d2 = ResidualBlock(512, [1, 3, 1], [256, 256, 1024], 6, stride=2)
        self.d3 = ResidualBlock(1024, [1, 3, 1], [512, 512, 2048], 3, stride=2)

        self.conv_bot = nn.Conv2d(2048, 1024, 1, stride=1, padding=0, bias=False)

        def create_decoder_branch(out_ch=2, ksize=5):
            module_list = [ 
                ("conva", nn.Conv2d(1024, 256, ksize, stride=1, padding=0, bias=False)),
                ("dense", DenseBlock(256, [1, ksize], [128, 32], 8, split=4)),
                ("convf", nn.Conv2d(512, 512, 1, stride=1, padding=0, bias=False),),
            ]
            u3 = nn.Sequential(OrderedDict(module_list))

            module_list = [ 
                ("conva", nn.Conv2d(512, 128, ksize, stride=1, padding=0, bias=False)),
                ("dense", DenseBlock(128, [1, ksize], [128, 32], 4, split=4)),
                ("convf", nn.Conv2d(256, 256, 1, stride=1, padding=0, bias=False),),
            ]
            u2 = nn.Sequential(OrderedDict(module_list))

            module_list = [ 
                ("conva/pad", TFSamepaddingLayer(ksize=ksize, stride=1)),
                ("conva", nn.Conv2d(256, 64, ksize, stride=1, padding=0, bias=False),),
            ]
            u1 = nn.Sequential(OrderedDict(module_list))

            module_list = [ 
                ("bn", nn.BatchNorm2d(64, eps=1e-5)),
                ("relu", nn.ReLU(inplace=True)),
                ("conv", nn.Conv2d(64, out_ch, 1, stride=1, padding=0, bias=True),),
            ]
            u0 = nn.Sequential(OrderedDict(module_list))

            decoder = nn.Sequential(
                OrderedDict([("u3", u3), ("u2", u2), ("u1", u1), ("u0", u0),])
            )
            return decoder

        ksize = 5 if mode == 'original' else 3
        if nr_types is None:
            self.decoder = nn.ModuleDict(
                OrderedDict(
                    [
                        ("np", create_decoder_branch(ksize=ksize,out_ch=2)),
                        ("hv", create_decoder_branch(ksize=ksize,out_ch=2)),
                    ]
                )
            )
        else:
            self.decoder = nn.ModuleDict(
                OrderedDict(
                    [
                        ("tp", create_decoder_branch(ksize=ksize, out_ch=nr_types)),
                        ("np", create_decoder_branch(ksize=ksize, out_ch=2)),
                        ("hv", create_decoder_branch(ksize=ksize, out_ch=2)),
                    ]
                )
            )

        self.upsample2x = UpSample2x()
        # TODO: pytorch still require the channel eventhough its ignored
        self.weights_init()

    def forward(self, imgs):

        imgs = imgs / 255.0  # to 0-1 range to match XY

        if self.training:
            d0 = self.conv0(imgs)
            d0 = self.d0(d0, self.freeze)
            with torch.set_grad_enabled(not self.freeze):
                d1 = self.d1(d0)
                d2 = self.d2(d1)
                d3 = self.d3(d2)
            d3 = self.conv_bot(d3)
            d = [d0, d1, d2, d3]
        else:
            d0 = self.conv0(imgs)
            d0 = self.d0(d0)
            d1 = self.d1(d0)
            d2 = self.d2(d1)
            d3 = self.d3(d2)
            d3 = self.conv_bot(d3)
            d = [d0, d1, d2, d3]

        # TODO: switch to `crop_to_shape` ?
        if self.mode == 'original':
            d[0] = crop_op(d[0], [184, 184])
            d[1] = crop_op(d[1], [72, 72])
        else:
            d[0] = crop_op(d[0], [92, 92])
            d[1] = crop_op(d[1], [36, 36])

        out_dict = OrderedDict()
        for branch_name, branch_desc in self.decoder.items():
            u3 = self.upsample2x(d[-1]) + d[-2]
            u3 = branch_desc[0](u3)

            u2 = self.upsample2x(u3) + d[-3]
            u2 = branch_desc[1](u2)

            u1 = self.upsample2x(u2) + d[-4]
            u1 = branch_desc[2](u1)

            u0 = branch_desc[3](u1)
            out_dict[branch_name] = u0

        return out_dict


####
def create_model(mode=None, **kwargs):
    if mode not in ['original', 'fast']:
        assert "Unknown Model Mode %s" % mode
    return HoVerNet(mode=mode, **kwargs)




"""
Pipeline for evaluation of models.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
from argparse import Namespace
import logging
from logging import Logger
import numpy as np
import pandas as pd
from . import eval_segment
import os
import shutil
from .postprocessing import join_hovprob_graph_main, join_graph_gt_main
from .preprocessing import hovernet2centroids_main, geojson2pngcsv_main, pngcsv2centroids_main, graph2centroids_main
from .utils.preprocessing import get_names
from .utils.pipes import HovernetNotFoundError, check_void
from .utils.classification import metrics_from_predictions
from .classification import infer_gnn
from .infer_pipe import set_best_configuration


def run_preprocessing(args: Namespace, logger: Logger) -> None:
    """
    Runs the preprocessing steps to convert the gson format to other formats.

    :param args: The arguments for the preprocessing.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger

    :raises HovernetNotFoundError: If the Hovernet centroids are not found in the training output.
    """
    logger.info('Extracting Hovernet centroids from training output.')
    if not os.path.isdir(os.path.join(args.root_dir, 'data', 'tmp_hov', 'json')):
        raise HovernetNotFoundError('Please, train again or extract hovernet outputs.')
    if check_void(os.path.join(args.root_dir, 'data', 'tmp_hov', 'centroids_hov')):
        newargs = Namespace(
            json_dir=os.path.join(args.root_dir, 'data', 'tmp_hov', 'json'),
            output_path=os.path.join(args.root_dir, 'data', 'tmp_hov', 'centroids_hov'),
        )
        hovernet2centroids_main(newargs)
    for split in ['train', 'validation', 'test']:
        if check_void(os.path.join(args.root_dir, 'data', split, 'names.txt')):
            logger.info(f'Preprocessing split {split}.')
            split_names = get_names(os.path.join(args.root_dir, 'data', split, 'gson'), '.geojson')
            with open(os.path.join(args.root_dir, 'data', split, 'names.txt'), 'w') as f:
                for name in split_names:
                    print(name, file=f)
        if check_void(os.path.join(args.root_dir, 'data', split, 'png')) or check_void(os.path.join(args.root_dir, 'data', split, 'csv')):
            logger.info('   From geojson to pngcsv.')
            newargs = Namespace(
                gson_dir=os.path.join(args.root_dir, 'data', split, 'gson'),
                png_dir=os.path.join(args.root_dir, 'data', split, 'png'),
                csv_dir=os.path.join(args.root_dir, 'data', split, 'csv'),
                num_classes=args.num_classes,
            )
            geojson2pngcsv_main(newargs)
        if check_void(os.path.join(args.root_dir, 'data', split, 'centroids')):
            logger.info('   Extracting centroids from GT.')
            newargs = Namespace(
                png_dir=os.path.join(args.root_dir, 'data', split, 'png'),
                csv_dir=os.path.join(args.root_dir, 'data', split, 'csv'),
                output_path=os.path.join(args.root_dir, 'data', split, 'centroids')
            )
            pngcsv2centroids_main(newargs)

    if args.best_arch is None:
        args.best_arch = 'GCN'
        set_best_configuration(args, logger)
    # Perform inference steps from infer_pipe
    if check_void(os.path.join(args.root_dir, 'tmp_graphs', 'gnn_preds')):
        logger.info('Starting graph inference.')
        os.makedirs(os.path.join(args.root_dir, 'tmp_graphs', 'gnn_preds'), exist_ok=True)
        os.makedirs(os.path.join(args.root_dir, 'tmp_graphs', 'graphs', 'raw'), exist_ok=True)
        os.makedirs(os.path.join(args.root_dir, 'tmp_graphs', 'graphs', 'hovpreds'), exist_ok=True)
        for split in ['train', 'validation', 'test']: # Move raw morphological features
            for file in os.listdir(os.path.join(args.root_dir, 'data', split, 'graphs', 'raw')):
                shutil.copy(
                    os.path.join(args.root_dir, 'data', split, 'graphs', 'raw', file),
                    os.path.join(args.root_dir, 'tmp_graphs', 'graphs', 'raw')
                )
        # Add hov probabilities
        newargs = Namespace(
            json_dir=os.path.join(args.root_dir, 'data', 'tmp_hov', 'json'),
            graph_dir=os.path.join(args.root_dir, 'tmp_graphs', 'graphs', 'raw'),
            output_dir=os.path.join(args.root_dir, 'tmp_graphs', 'graphs', 'hovpreds'),
            num_classes=args.num_classes,
            enable_background=args.enable_background,
        )
        join_hovprob_graph_main(newargs, logger)
        for file in os.listdir(os.path.join(args.root_dir, 'tmp_graphs', 'graphs', 'hovpreds')):
            df = pd.read_csv(os.path.join(os.path.join(args.root_dir, 'tmp_graphs', 'graphs', 'hovpreds'), file))
            assert (df['class'].to_numpy() == ((df['prob1'].to_numpy() > 0.5) * 1 + 1)).all(), f"Probs and prediction of Hovernet don't coincide. {file}"
            df.drop('class', axis=1).to_csv(os.path.join(os.path.join(args.root_dir, 'tmp_graphs', 'graphs', 'hovpreds'), file), index=False)
        model_name = 'best_' + args.best_arch + '_' + args.best_num_layers + '_' \
            + args.best_dropout + '_' + args.best_norm_type
        if args.gnn_dir is None:
            args.gnn_dir = os.path.join(args.root_dir, 'weights', 'classification', 'gnn')
        newargs = Namespace(
            node_dir=os.path.join(args.root_dir, 'tmp_graphs', 'graphs', 'hovpreds'),
            output_dir=os.path.join(args.root_dir, 'tmp_graphs', 'gnn_preds'),
            weights=os.path.join(args.gnn_dir, 'weights', model_name + '.pth'),
            conf=os.path.join(args.gnn_dir, 'confs', model_name + '.json'),
            normalizers=os.path.join(args.gnn_dir, 'normalizers', model_name + '.pkl'),
            num_classes=args.num_classes,
            disable_prior=args.disable_prior,
            disable_morph_feats=args.disable_morph,
            enable_background=args.enable_background,
        )
        infer_gnn(newargs)
    # Postproc to obtain centroids
    if check_void(os.path.join(args.root_dir, 'tmp_graphs', 'centroids')):
        logger.info('Parsing gnn output.')
        logger.info('   Converting .nodes.csv to .centroids.csv.')
        newargs = Namespace(
            graph_dir=os.path.join(args.root_dir, 'tmp_graphs', 'gnn_preds'),
            centroids_dir=os.path.join(args.root_dir, 'tmp_graphs', 'centroids'),
            num_classes=args.num_classes,
            enable_background=args.enable_background,
        )
        graph2centroids_main(newargs)

    logger.info('Creating folders.')
    os.makedirs(os.path.join(args.save_dir, 'conf-matrices', 'gnn_individual'), exist_ok=True)
    os.makedirs(os.path.join(args.save_dir, 'conf-matrices', 'hov_individual'), exist_ok=True)
    for split in ['train', 'validation', 'test']:
        os.makedirs(os.path.join(args.save_dir, split), exist_ok=True)
    return


def compute_ece(args: Namespace, split: str, is_hov: bool) -> None:
    """
    Computes Expected Calibration Error (ECE) and other metrics for the given split.

    :param args: The arguments for computing ECE and metrics.
    :type args: Namespace

    :param split: The name of the split (e.g., 'train', 'validation', 'test').
    :type split: str

    :param is_hov: Whether we are computing for hovernet or gnns. The directories have different structure, so we need to know.
    :type is_hov: bool
    """
    if is_hov:
        folder = os.path.join(args.root_dir, 'data', split, 'graphs', 'preds')
    else:
        folder = os.path.join(args.root_dir, 'tmp_graphs', 'gnn_preds_gt', split)
    trues, probs = None, None
    for file in os.listdir(folder):
        df = pd.read_csv(os.path.join(folder, file))
        _trues = df['class'].to_numpy() - 1
        if trues is None:
            trues = _trues.reshape((-1, 1))
        else:
            trues = np.vstack((trues, _trues.reshape((-1, 1))))
        if args.num_classes == 2:
            _probs = df['prob1'].to_numpy().reshape((-1, 1))
            _probs = np.hstack((1 - _probs, _probs))
        else:
            cols = ['prob' + str(k) for k in range(1, args.num_classes + 1)]
            _probs = df[cols].to_numpy()
        if probs is None:
            probs = _probs
        else:
            probs = np.vstack((probs, _probs))
        if pd.isna(trues).any():
            print(file)
    preds = np.argmax(probs, axis=1).reshape((-1, 1))
    metrics = metrics_from_predictions(trues, preds, probs[:, 1], args.num_classes)
    if args.num_classes == 2:
        acc, f1, auc, perc_error, ece = metrics
        dic_metrics = {
            'F1': [f1], 'Accuracy': [acc], 'ROC_AUC': [auc], 'Perc_err': [perc_error], 'ECE': [ece]
        }
    else:
        micro, macro, weighted, ece = metrics
        dic_metrics = {
            'Macro F1': [macro], 'Weighted F1': [weighted], 'Micro F1': [micro], 'ECE': [ece]
        }
    metrics_df = pd.DataFrame(dic_metrics)
    metrics_df.to_csv(os.path.join(args.save_dir, split, ('hov' if is_hov else 'gnn') + '_ece.csv'), index=False)


def run_evaluation(args: Namespace, logger: Logger) -> None:
    """
    Runs the evaluation of Hovernet output for different splits.

    :param args: The arguments for the evaluation.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Starting evaluation of Hovernet output.')
    for split in ['train', 'validation', 'test']:
        logger.info(f'    Evaluating {split} split')
        newargs = Namespace(
            names=os.path.join(args.root_dir, 'data', split, 'names.txt'),
            gt_path=os.path.join(args.root_dir, 'data', split, 'centroids'),
            pred_path=os.path.join(args.root_dir, 'data', 'tmp_hov', 'centroids_hov'),
            save_name=os.path.join(args.save_dir, split, 'hov'),
            debug_path=os.path.join(args.save_dir, 'conf-matrices', 'hov_individual', 'debug_hov') if args.debug else None,
            num_classes=args.num_classes,
        )
        eval_segment(newargs, logger)
        compute_ece(args, split, is_hov=True)
        if args.debug:
            shutil.move(
                os.path.join(args.save_dir, 'conf-matrices', 'hov_individual', 'debug_hov_global.csv'),
                os.path.join(args.save_dir, 'conf-matrices', 'debug_hov_global_' + split + '.csv'))
    # Evaluate graph output same as hov output
    logger.info('Starting evaluation of GNN output.')
    for split in ['train', 'validation', 'test']:
        logger.info(f'    Evaluating {split} split (with background)')
        newargs = Namespace(
            names=os.path.join(args.root_dir, 'data', split, 'names.txt'),
            gt_path=os.path.join(args.root_dir, 'data', split, 'centroids'),
            pred_path=os.path.join(args.root_dir, 'tmp_graphs', 'centroids'),
            save_name=os.path.join(args.save_dir, split, 'gnn'),
            debug_path=os.path.join(args.save_dir, 'conf-matrices', 'gnn_individual', 'debug_gnn') if args.debug else None,
            num_classes=args.num_classes
        )
        eval_segment(newargs, logger)
        if args.debug:
            shutil.move(
                os.path.join(args.save_dir, 'conf-matrices', 'gnn_individual', 'debug_gnn_global.csv'),
                os.path.join(args.save_dir, 'conf-matrices', 'debug_gnn_global_' + split + '.csv'))
    for split in ['train', 'validation', 'test']:
        os.makedirs(os.path.join(args.root_dir, 'tmp_graphs', 'gnn_preds_gt', split), exist_ok=True)
        for file in get_names(os.path.join(args.root_dir, 'data', split, 'graphs', 'preds'), '.nodes.csv'):
            df_gt = pd.read_csv(os.path.join(args.root_dir, 'data', split, 'graphs', 'preds', file + '.nodes.csv'))
            df = pd.read_csv(os.path.join(args.root_dir, 'tmp_graphs', 'gnn_preds', file + '.nodes.csv'))
            df = df[df.id.isin(df_gt.id)]
            df = df.set_index(df.id)
            df_gt = df_gt.set_index(df_gt.id)
            df['class'] = df_gt['class']
            assert not df['class'].isna().any()
            df.to_csv(os.path.join(args.root_dir, 'tmp_graphs', 'gnn_preds_gt', split, file + '.nodes.csv'), index=False)
        compute_ece(args, split, is_hov=False)
    shutil.rmtree(os.path.join(args.root_dir, 'tmp_graphs'))
    return


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--root-dir', type=str, default='./.internals/', help='Root folder to save data and models.')
    parser.add_argument('--save-dir', type=str, required=True, help='Folder to save the results, without file type.')
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    parser.add_argument('--best-num-layers', type=str, help='Optimal number of layers when training GNNs.')
    parser.add_argument('--best-dropout', type=str, help='Optimal dropout rate when training GNNs')
    parser.add_argument('--best-norm-type', type=str, help='Optimal type of normalization layers when training GNNs')
    parser.add_argument('--best-arch', type=str, help='Best architecture (convolutional, attention, ...) when training GNNs', choices=['GCN', 'ATT'])
    parser.add_argument('--debug', action='store_true', help='Whether to save confusion matrices.')
    parser.add_argument('--gnn-dir', type=str, help='Path where the gnn is saved, otherwise the default in root-dir will be used. It must have three folder inside: confs, weights and normalizers.')
    parser.add_argument('--disable-prior', action='store_true', help='Whether to disable the use of Hovernet probabilities.')
    parser.add_argument('--disable-morph', action='store_true', help='Whether to disable the use of morphological properties.')
    parser.add_argument('--enable-background', action='store_true', help='If enabled, GNNs are allowed to predict the class 0 (background) and correct extra cells.')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()

    logger = logging.getLogger('eval_pipe')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    run_preprocessing(args, logger)
    run_evaluation(args, logger)



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os
import pandas as pd
import numpy as np

from tumourkit.preprocessing.centroidspng2csv import centroidspng2csv
from tumourkit.utils.preprocessing import get_names, parse_path, read_png, read_csv, read_centroids

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PNGCSV_DIR = parse_path(TEST_DIR) + 'pngcsv/'
THRESHOLD = 0.1


def check_equal(a: pd.DataFrame, b: pd.DataFrame) -> bool:
    return np.abs(a.to_numpy() - b.to_numpy()).mean() < THRESHOLD


@pytest.mark.parametrize("name", get_names(PNGCSV_DIR, '.centroids.csv'))
def test_centroidspng2csv(name):
    centroids_file = read_centroids(name, PNGCSV_DIR)
    png_file = read_png(name, PNGCSV_DIR)
    csv_file = centroidspng2csv(centroids_file, png_file)
    ref_csv = read_csv(name, PNGCSV_DIR)
    if not check_equal(csv_file, ref_csv):
        assert False, name
    assert(True)



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os

from tumourkit.utils.preprocessing import parse_path
from tumourkit.utils.postprocessing import get_greatest_connected_component

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
RSWOOSH_DIR = parse_path(TEST_DIR) + 'rswoosh/'
THRESHOLD = 0.0001

@pytest.mark.parametrize("ids,max_id,out_l,out_r", [
    ([1,2,3,4,6,7,8], 8, 1, 4),
    ([1,4,3,7,8,6,2], 8, 1, 4),
    ([0,2,3,4,5,6,9,11], 11, 2, 6),
    ([3,2,0,4,11,5,6,9], 11, 2, 6),
    ([9,10,0,1,2,6,8], 10, 8, 2),
    ([0,1,2,3,4,5,6], 10, 0, 6),
    ([0,1,2,3,4,5,6], 6, 0, 6),
    ([0,1,2,3,4,5,6,8,9,10], 11, 0, 6),
    ([0,1,2,3,4,5,6,8,9,10], 10, 8, 6)
])
def test_conversion(ids, max_id, out_l, out_r):
    l, r = get_greatest_connected_component(ids, max_id)
    assert(l == out_l and r == out_r)
